package clarion.acs;
import clarion.system.*;
import clarion.common.*;
import clarion.acs.net.*;
import clarion.acs.net.auxiliary.*;
import clarion.acs.gs.*;
import clarion.acs.wm.*;
import clarion.nacs.*;

public class ACS {

   // global settings.
   Global  global;
   // task-specific clarion settings.
   TaskClarion taskClarion;
   // eligibility of the external networks.
   short[][][][] eligCondions;
   EligibilityCheck eligCheck;
   boolean emOn;
   double  emTrainAcsProb;
   boolean aemOn;
   double  aemTrainAcsProb;
   // current sensory input, wm content and goal.
   Feature sensoryInput[];
   Feature desiredOutput[][];
   // variables for ACS components.
   int netNum, exNetNum, activeCompNum;
   int exNetIdx, wmNetIdx, gsNetIdx;
   int gsOption;
   boolean wmOnFlag;
   short[] activeComps[];
   AcsNet  nets[];
   // variables for current state.
   double[] compStates[];
   int  corExtMode;       // coordinating multiple external action net.
   int  actionTypeNum;    // # of different acitons: GS, WM, orvaried EX
   int  chosenType;       // types: BL, RER, IRL, FR. used for perceentage mode.
   // current chosen actions, actually is indices of actions in one action set.
   // the actual action is a dim-value pairs.
   short[]  chosenActions[];  // action types: EX, WM, or GS, for stochastic mode.
   // retrieval results from NACS.
   GKSChunk retrievalResults[];

   public ACS( Global g )
   {
       global = g;
       global.setACS( this );
       taskClarion = global.getTaskClarion();
       // initialize eligibility of external networks.
       exNetNum = global.getExNetNum();
       eligCondions = new short[exNetNum][][][];
       for( int i = 0; i < exNetNum; i++ )
            eligCondions[i] = taskClarion.getEligibility(i);
       eligCheck = new EligibilityCheck( global, eligCondions );
       emOn = global.isEMon();
       if( emOn )
           emTrainAcsProb = global.getEmTrainProb();
       aemOn = global.isAEMon();
       if( aemOn )
           aemTrainAcsProb = global.getAemTrainAcsProb();
       // initialize overall nets.
       netNum = global.getAcsNetNum();
       wmOnFlag = global.isWmOn();
       gsOption = global.getGsOption();
       wmNetIdx = global.getWmNetIdx();
       gsNetIdx = global.getGsNetIdx();
       nets = new AcsNet[ netNum ];
       corExtMode = global.getCorMultiExtMode();
       if( corExtMode == Global.COR_EXT_ALL ){
           activeCompNum = netNum;
           activeComps = new short[ activeCompNum ][ 2 ];
           for( int i = 0; i < exNetNum; i++ ){
                activeComps[ i ][ 0 ] = (short)Global.EX_TYPE;
                activeComps[ i ][ 1 ] = (short)i;
           }
           if( wmOnFlag ){
               activeComps[ wmNetIdx ][ 0 ] = (short)Global.WM_TYPE;
               activeComps[ wmNetIdx ][ 1 ] =(short)wmNetIdx;
           }
           if( gsOption != global.GS_NON ){
               activeComps[ gsNetIdx ][ 0 ] = (short)Global.GS_TYPE;
               activeComps[ gsNetIdx ][ 1 ] = (short)gsNetIdx;
           }
       }
       else   // SINGLE_MODE.
       {
           activeCompNum = 1;
           if( wmOnFlag )
               activeCompNum++;
           if( gsOption != global.GS_NON )
               activeCompNum++;
           activeComps = new short[ activeCompNum ][ 2 ];
           activeComps[ 0 ][ 0 ] = (short)Global.EX_TYPE;
           activeComps[ 0 ][ 1 ] = 0;
           if( wmOnFlag ){
               activeComps[ 1 ][ 0 ] = (short)Global.WM_TYPE;
               activeComps[ 1 ][ 1 ] = (short)wmNetIdx;
           }
           if( gsOption != global.GS_NON ){
               if( wmOnFlag ){
                   activeComps[ 2 ][ 0 ] = (short)Global.GS_TYPE;
                   activeComps[ 2 ][ 1 ] = (short)gsNetIdx;
               }
               else{
                   activeComps[ 1 ][ 0 ] = (short)Global.GS_TYPE;
                   activeComps[ 1 ][ 1 ] = (short)gsNetIdx;
               }
           }
       }
       chosenActions = new short[activeCompNum][];
       compStates = new double[ netNum ][];
       for( int i = 0; i < netNum; i++ )
            compStates[ i ] = new double[ global.getIdnInputNum(i) ];
       // initialize External component.
       for( int i = 0; i < exNetNum; i++ )
            nets[ i ] = new EXNet( i, global );
       // initialize WM component if exist.
       if( wmOnFlag )
           nets[ wmNetIdx ] = new WMNet( wmNetIdx, global );
       // initialize GS component if exist.
       if( gsOption != global.GS_NON )  // a goal structure is selected.
           nets[ gsNetIdx ] = new GSNet( gsNetIdx, global );
   }

   public int getSensoryLength()
   {
      return global.getSensoryInputLength();
   }

   public int getGsLength()
   {
      if( global.getGsOption() != Global.GS_NON )
          return global.getIdnInputNum( gsNetIdx );
      return 0;
   }

   public int getWmLength()
   {
      if( global.isWmOn() )
          return global.getIdnInputNum( wmNetIdx );
      return 0;
   }

   public String[] getRerRules( int netIdx )
   {
      if( nets[ netIdx ] != null )
          return nets[ netIdx ].getRerRules();
      else
          return (String[])null;
   }

   public String[] getIrlRules( int netIdx )
   {
      if( nets[ netIdx ] != null )
          return nets[ netIdx ].getIrlRules();
      else
          return (String[])null;
   }

   public String[] getFixRules( int netIdx )
   {
      if( nets[ netIdx ] != null )
          return nets[ netIdx ].getFixRules();
      else
          return (String[])null;
   }

   public void getChosenActions( short[][] arr )
   {
      for( int i = 0; i < activeCompNum; i++ )
        for( int j = 0; j < chosenActions[i].length; j++ ){
             int idx = activeComps[i][1];
             arr[idx][j] = chosenActions[i][j];
        }
   }

   public Chunk[][][] getAllConclusions()
   {
      Chunk[][][] chunks = new Chunk[activeCompNum][][];
      for( int i = 0; i < activeCompNum; i++ )
           chunks[i] = nets[ activeComps[i][1] ].getAllConclusions();
      return chunks;
   }

   public int getRT()
   {
      int RT = 0;
      for( int i = 0; i < activeCompNum; i++ )
           if( RT < nets[ activeComps[i][1] ].getRT() )
               RT = nets[ activeComps[i][1] ].getRT();
      return RT;
   }

   public int getRT( int nacsRT )
   {
      int RT = 0;
      for( int i = 0; i < activeCompNum; i++ )
           if( RT < nets[ activeComps[i][1] ].getRT( nacsRT ) )
               RT = nets[ activeComps[i][1] ].getRT( nacsRT );
      return RT;
   }

   public void setSensoryInput( Feature[] input )
   {
      sensoryInput = input;
   }

   public void setCurDesiredOutput( Feature[][] output )
   {
      desiredOutput = output;
   }

   public void setRetrievalResult( GKSChunk[] result, int slot )
   {
      retrievalResults = result;
      if( wmOnFlag )
          ((WMNet)nets[ wmNetIdx ]).setRetrievalResult( retrievalResults, slot );
   }

   public void setCurrentTime( long curTime )
   {
      for( int i = 0; i < nets.length; i++ )
           nets[ i ].setCurrentTime( curTime );
   }

   // observe new sensory input, new GS and WM states if exists and deliver them to each
   // active network and form new state including the three types of inputs if exists.
   public void observeState( Feature[] sInput )
   {
      // receive sensory input.
      setSensoryInput( sInput );
      // check which external net will be used.
      if( corExtMode == Global.COR_EXT_ONE ){
          exNetIdx = eligCheck.getEligibleNetwork( sInput );
          activeComps[ 0 ][ 1 ] = (short)exNetIdx;
      }
      // deliver sensory, GS and WM input to each eligible net if exists.
      for( int i = 0; i < activeCompNum; i++ ){
           nets[ activeComps[i][1] ].setSensoryInput( sensoryInput );
           if( activeComps[ i ][ 0 ] == Global.GS_TYPE ){
               nets[ activeComps[i][1] ].getGsState( compStates[ activeComps[i][1] ] );
               for( int j = 0; j < activeCompNum; j++ )
                    nets[ activeComps[j][1] ].setGsState( compStates[ activeComps[i][1] ] );
           }
           if( activeComps[ i ][ 0 ] == Global.WM_TYPE ){
               nets[ activeComps[i][1] ].getWmState( compStates[ activeComps[i][1] ] );
               for( int j = 0; j < activeCompNum; j++ )
                    nets[ activeComps[j][1] ].setWmState( compStates[ activeComps[i][1] ] );
           }
      }
      // form current state.
      for( int i = 0; i < activeCompNum; i++ )
          nets[ activeComps[i][1] ].formState();
   }

   // after the new state (input), including EX, GS, WM if exists, is given, this method is
   // called. decides an action of each active network and gets them.
   public void reasoning()
   {
      // let each active component reasoning.
      for( int i = 0; i < activeCompNum; i++ ){
           nets[ activeComps[i][1] ].reasoning();
           chosenActions[ activeComps[i][1] ] = nets[ activeComps[i][1] ].getChosenAction();
      }
   }

   public void performWmAction()
   {
      nets[ wmNetIdx ].perform();
   }

   public void performGsAction()
   {
      nets[ gsNetIdx ].perform();
   }

   // update each active network. update GS, WM states if exists by performing the chosen
   // action. the EX action is already performed before this method is called.
   public void update( Feature[] sInput, double[] reinforcements )
   {
      // let each acs component update.
      for( int i = 0; i < activeCompNum; i++ ){
           int netIdx = activeComps[i][1];
           if( global.getIdnLearningOption( netIdx ) == Global.FIX )
               continue;
           if( activeComps[i][0] == global.EX_TYPE && global.getIdnLearningOption( netIdx ) == global.BP ){
               if( desiredOutput == null || desiredOutput[netIdx] == null )
                   System.out.println( "No desired output to train the IDN! (in ACS.java)" );
               else
                   nets[ netIdx ].setCurDesiredOutput( desiredOutput[netIdx] );
           }
           if( global.getIdnLearningOption( netIdx ) == global.QL )
                observeState( sInput );
           nets[ netIdx ].update( reinforcements[ netIdx ] );
      }
      // offline training ACS usning EM.
      if( emOn && Math.random() < emTrainAcsProb )
          offlineTrainByEm();
      // offline training ACS usning AEM.
      if( aemOn && Math.random() < aemTrainAcsProb )
          offlineTrainByAem();
   }

/******************************* PRIVATE METHOD *******************************/
  private void offlineTrainByEm()
  {
      for( int i = 0; i < activeCompNum; i++ )
           nets[ activeComps[i][0] ].offlineTrainByEm();
  }

  private void offlineTrainByAem()
  {
      for( int i = 0; i < activeCompNum; i++ )
           nets[ activeComps[i][0] ].offlineTrainByAem();
  }
}