// implements Q learning over a backprop neural net.
// input is a state, and output is Q values of each action
package clarion.common;
import clarion.system.*;

public class QBPNet extends BPNet {

    // global settings.
    Global global;
    // net index in ACS.
    int netIdx;
    // variables for the net.
    protected int  actionDimNum;
    protected short  actionPerformed[]; // each action with multiple dims.
    protected double discount, reinforcement;
    protected double state[], preState[];

    public QBPNet( int inputNum, int nIdx, Global g )
    {
       	super( inputNum, nIdx, g, Global.ACS );
        netIdx = nIdx;
        global = g;
        actionDimNum = global.getIdnOutputDimNum( netIdx );
        actionPerformed = new short[ actionDimNum ];
        discount = global.getIdnQDiscountRate( netIdx );
       	state = new double[ nInput ];
        if( global.getIdnLearningOption( netIdx ) == global.QL )
            preState = new double[ nInput ];
    }

    public void reinit()
    {
        resetMatches();
        reinitWeights();
    }

    public void getState( int[] fillMe )
    {
    	for(int i = 0; i < nInput; i++ )
             fillMe[i] = (int) state[i];
    }

    public void setState( int[] curState )
    {
        if( preState != null )  // normal Q-Learning.
        {
            for( int i = 0; i < nInput; i++ )
                 preState[ i ] = state[ i ];
        }
        for( int i = 0; i < nInput; i++ )
             state[ i ] = curState[ i ];

        setInput();
    }

    public double getQDiscount()
    {
        return discount;
    }

    public double getReinforcement()
    {
        return reinforcement;
    }

    public void setState( double[] curState )
    {
        if( preState != null )  // normal Q-Learning.
        {
            for( int i = 0; i < nInput; i++ )
                 preState[ i ] = state[ i ];
        }
        for( int i = 0; i < nInput; i++ )
             state[ i ] = curState[ i ];

        setInput();
    }

    // used for normal Q-Learning.
    public void setPreState()
    {
        for( int i = 0; i < nInput; i++ )
             state[ i ] = preState[ i ];

        setInput();
    }

    public void setInput()
    {
	setInput( state );
    }

    public void setReinforcement( double value )
    {
	reinforcement = value;
    }

    public double getQVal( int index )
    {
       return Output[ index ];
    }

    public void getQVals( double[] FillMe )
    {
    	for (int i = 0; i < nOutput; i++ )
              FillMe[i] = Output[i];
    }

    // new method for action with multiple dims.
    public void getQVals( short[] action, double[] FillMe )
    {
       	for( int i = 0; i < actionDimNum; i++ )
             FillMe[i] = getOutput( i, action[i] );
    }

   // to find the max Q value of all the Q values.
    public double getMaxQVal()
    {
        double highest = Output[ 0 ];
        for( int i = 1; i < nOutput; i++ )
             if( Output[ i ] > highest )
                 highest = Output[ i ];
        return highest;
    }

    // new method for action with multiple dims.
    public void getMaxQVals( double[] FillMe )
    {
        for( int i = 0; i < actionDimNum; i++ ) {
             double highest = Output[ outputOffsets[i] ];
             for( int j = 1; j < outputDVs[i]; j++ )
                  if( Output[ outputOffsets[i] + j ] > highest )
                      highest = Output[ outputOffsets[i] + j ];
             FillMe[ i ] = highest;
        }
    }

    // let the system know what action was performed
    public void setPerformedAction( short[] action )
    {
        for( int i = 0; i < actionDimNum; i++ )
             actionPerformed[ i ] = action[ i ];
    }

    // used for simplified Q-Learning.
    public void update( short[] action, double feedback, double[] qValue )
    {
	       //let QBP know what action was performed
       	setPerformedAction( action );
        // update the reinforcement
        setReinforcement( feedback );
        // update qbp net
       	backwardPass( qValue );
    }

    // used for normal Q-Learning.
    public void update( short[] action, double feedback, double[] qVal1, double[] qVal2 )
    {
       	//let QBP know what action was performed
       	setPerformedAction( action );
        //update the reinforcement
        setReinforcement( feedback );
        // update qbp net
        setPreState();
        forwardPass();
       	backwardPass( qVal1, qVal2 );
    }

/************************** QBP version****************************************/
    public void backwardPass( double[] qValue1, double[] qValue2 )
    {
       	computeErrors( qValue1, qValue2 );
        computeHiddenErrors();
        modifyHiddenToOutput();
        modifyInputToHidden();
    }

    public void backwardPass( double[] qValue )
    {
       	computeErrors( qValue );
        computeHiddenErrors();
        modifyHiddenToOutput();
        modifyInputToHidden();
    }

/************************** PROTECTED METHODS *********************************/
   // compute the errors of QBP network
    protected void computeErrors( double[] qValue1, double[] qValue2 )
    {
        for( int i = 0; i < actionDimNum; i++ )
         for( int j = 0; j < outputDVs[i]; j++ ){
              if( j != actionPerformed[i] )
                  Errors[ outputOffsets[i] + j ] = 0;
              else{
	          double err = reinforcement + discount * qValue2[i] - qValue1[i];
       		  Errors[ outputOffsets[i] + j ] = OutputDeriv[ outputOffsets[i] + j ] * err;
                                            sumSqErrors = err * err / 2;
              }
        }
    }

    // compute the errors of simplified QBP network
    protected void computeErrors( double[] qValue )
    {
        for( int i = 0; i < actionDimNum; i++ )
         for( int j = 0; j < outputDVs[i]; j++ ){
              if( j != actionPerformed[i] )
                  Errors[ outputOffsets[i] + j ] = 0;
              else{
	          double err = reinforcement - qValue[i];
       		  Errors[ outputOffsets[i] + j ] = OutputDeriv[ outputOffsets[i] + j ] * err;
                                            sumSqErrors = err * err / 2;
              }
        }
    }

    // optimized, since all errors except for one are 0
    protected void modifyHiddenToOutput()
    {
        for( int j = 0; j < actionDimNum; j++ ) {
          int idx = outputOffsets[ j ] + actionPerformed[ j ];
          double x = Errors[ idx ];
          for( int i = 0; i < nHidden; i++ ) {
             // update the HiddenToOutputWeights
             if( Momentum > RZero ) {
                 /* momentum */
                 HiddenToOutputMomentum[ i ] [ idx ] = Eta * x * Hidden[ i ]
                           + Momentum * HiddenToOutputMomentum[ i ][ idx ];
                 HiddenToOutputWeights[ i ][ idx ] +=
                           HiddenToOutputMomentum[ i ][ idx ];
             }
             else
                 HiddenToOutputWeights[i][idx] += Eta * x * Hidden[i];
          }
          // update the OutputThresholds
          if( Momentum > RZero ) {
              OutputMomentum[ idx ] =  Eta * x + Momentum * OutputMomentum[ idx ];
              OutputThresholds[ idx ] += OutputMomentum[ idx ];
          }
          else
              OutputThresholds[ idx ] += Eta * x;
        }
    }
}
