package clarion.nacs;
import java.lang.*;
import java.util.Vector;
import java.awt.*;
import clarion.tools.*;

public class GKNet {

  final int SCALE_VAL = 100;
  final short EMPTY_LINK = -1;

  // the reasoning methods.
  final int RULE_BASED = 0;
  final int SIMI_BASED = 1;
  final int RULE_SIMI_BASED = 2;

  int MAX_NODE_NUM;   // maxium # of nodes in GKN.
  int MAX_NODE_AGE;
  int MAX_LINK_AGE;   // used for keeping a node or link.
  int DIM_NUM;        // the # of dimensions of each node.
  int curNodeNum;     // current # of nodes in the net.
  int reasoningType;  // used reasoning method.
  int chosenNodeIdx;  // the index of chosen node by using the node decider.
  int emptyLinkIdx;   // the index of current empty link that can be used.
  int dimValNums[];   // the # of values in each dimension.
  int dimTypes[];     // the type of each dimension, 0: discrete; 1:continuous.
  int curInput[][];   // current input.

  short cueDimNum;
  short cueDimIndices[];

  int nodeAges[];           // ages of nodes.
  int fromLinkAges[][][];   // ages of links.
  int toLinkAges[][][];     // ages of links.

  // the reason for using the following variables in stead of using adjacent matrices
  //( two dimensional ) is that the associative links is very sparse in most case and
  // it is fairly time-consuming for searching those chunks a chunk connects to and
  // is connected with( the time complexity is O( N ), however, using 3-dimensional
  // ( the 2nd dimension of adjacent matrix is changed into a 2-dimensional array
  // ( like a square ), this can reduce the time complexity of the above searching
  // to O( sqrt( N ) ).
  short ROW_NUM;        // used for speeding search.
  short COLUMN_NUM;     // used for speeding search.
  short fromLinkNums[][];   // used for speeding.
  short toLinkNums[][];     // used for speeding.
  short fromLinks[][][];    // all the 'from' links in GKN.
  short toLinks[][][];      // all the 'to' links in GKN.

  short activatedNodeNum;
  short typeActivatedNodeNums[];
  short activatedNodeIndices[];
  short typeActivatedNodeIndices[][];

  short conditionChunkIdx;
  short conclusionChunkIdx;
  short concChunkIndices[];          // the indices for all conclusion chunks.
  short concDimIdx;
  short withinReasoningResults[];    // the results of reasoning.

  long  currentTime;        // used for calculating BLA.

  float weights[];
  double C14, C15;             // used for intermixed reasoning type.
  double ACTIVATE_THRESHOLD; // used for reasoning.
  double STRENTH_THRESHOLD;    // used for between nodes reasoning.
  double SIMI_THRESHOLD;       // used for similarity reasoning type.
  double INIT_BLA;             // used for calculating BLA.

  double typeActivations[][];  // activation of each node with different reasoning types
                               // ( within reasoning ).
  double activations[];        // activation of each node combining different reasoning types.
                               // ( within reasoning ).
  double typeStrengths[][];    // strength of conclusion chunk of associative rules
                               // with different reasoning types( between reasoning )
  double strengths[];          // strengths of conclusion chunks of associative rules
                               // combining different reasoning types( between reasoning )

  Vector availableNodeList; // indicate current active nodes.
  Vector emptyNodeList;     // indicate current empty nodes.
  Vector cue;               // cue used for retrieval.
  GKChunk nodes[];            // all the chunk nodes in GKN.
  StochasticDecider nodeDecider;   // used for selecting the activated node.

  // constructor.
  public GKNet( int nNum1, int nNum2, int dNum, int rType, int[] dValNums,
                int[] dTypes, double initBla )
  {
     ROW_NUM = (short)nNum1;
     COLUMN_NUM = (short)nNum2;
     MAX_NODE_NUM = nNum1 * nNum2;
     DIM_NUM = dNum;
     reasoningType = rType;
     INIT_BLA = initBla;

     if( reasoningType == RULE_BASED )
     {
         C14 = 1.0;
         C15 = 0.0;
     }

     if( reasoningType == SIMI_BASED )
     {
         C14 = 0.0;
         C15 = 1.0;
     }

     dimValNums = new int[ DIM_NUM ];
     curInput = new int[ DIM_NUM ][];
     for( int i = 0; i < DIM_NUM; i++ )
     {
          dimValNums[ i ] = dValNums[ i ];
          curInput[ i ] = new int[ dValNums[i] ];
     }

     dimTypes = new int[ DIM_NUM ];
     for( int i = 0; i < DIM_NUM; i++ )
          dimTypes[ i ] = dTypes[ i ];

     weights = new float[ DIM_NUM ];
     for( int i = 0; i < DIM_NUM; i++ )
          weights[ i ] = (float)( 1.0 / DIM_NUM );

     nodeAges = new int[ MAX_NODE_NUM ];
     fromLinkNums = new short[ MAX_NODE_NUM ][ ROW_NUM ];
     toLinkNums = new short[ MAX_NODE_NUM ][ ROW_NUM ];
     fromLinkAges = new int[ MAX_NODE_NUM ][ ROW_NUM ][ COLUMN_NUM ];
     toLinkAges = new int[ MAX_NODE_NUM ][ ROW_NUM ][ COLUMN_NUM ];
     fromLinks = new short[ MAX_NODE_NUM ][ ROW_NUM ][ COLUMN_NUM ];
     toLinks = new short[ MAX_NODE_NUM ][ ROW_NUM ][ COLUMN_NUM ];

     typeActivatedNodeNums = new short[ 2 ];
     activatedNodeIndices = new short[ MAX_NODE_NUM ];
     typeActivatedNodeIndices = new short[ MAX_NODE_NUM ][ 2 ];

     typeStrengths = new double[ MAX_NODE_NUM ][ 2 ];
     strengths = new double[ MAX_NODE_NUM ];
     typeActivations = new double[ MAX_NODE_NUM ][ 2 ];
     activations = new double[ MAX_NODE_NUM ];

     nodes = new GKChunk[ MAX_NODE_NUM ];
     availableNodeList = new Vector( MAX_NODE_NUM );
     emptyNodeList = new Vector( MAX_NODE_NUM );
     initNodeLists();
     for( int i = 0; i < MAX_NODE_NUM; i++ )
          initNodeVariables( i );

     cue = new Vector( DIM_NUM );
    // nodeDecider = new StochasticDecider( World.TASK_PARA_ARRAY[World.TEMPERATURE] );
     nodeDecider = new StochasticDecider( 0.01 );
  }

/**************************** INITINALIZATION METHODS ******************************/

  // initialize the variables relevant to chunks.
  public void initNodeVariables( int nodeIdx )
  {
       nodeAges[ nodeIdx ] = 0;
       for( int j = 0; j < ROW_NUM; j++ )
       {
            fromLinkNums[ nodeIdx ][ j ] = 0;
            toLinkNums[ nodeIdx ][ j ] = 0;
            for( int k = 0; k < COLUMN_NUM; k++ )
            {
                 fromLinkAges[ nodeIdx ][ j ][ k ] = 0;
                 toLinkAges[ nodeIdx ][ j ][ k ] = 0;
                 fromLinks[ nodeIdx ][ j ][ k ] = EMPTY_LINK;
                 toLinks[ nodeIdx ][ j ][ k ] = EMPTY_LINK;
            }
      }

      for( short i = 0; i < MAX_NODE_NUM; i++ )
      {
           if( nodes[ i ] != null )
           {
               removeLink( i, (short)nodeIdx );
               removeLink( (short)nodeIdx, i );
           }
      }
  }

  // initialize the list of existing chunks and the list of empty chunks for future use.
  public void initNodeLists()
  {
     availableNodeList.removeAllElements();
     emptyNodeList.removeAllElements();
     for( int i = 0; i < MAX_NODE_NUM; i++ )
     {
         Integer elem = new Integer( i );
         emptyNodeList.addElement( elem );
     }
  }

  // initialize the retrieval results.
  public void initResults()
  {
     withinReasoningResults = new short[ dimValNums[concDimIdx] ];
  }

  // reinitialize the net.
  public void reinit()
  {
     for( int i = 0; i < DIM_NUM; i++ )
       for( int j = 0; j < dimValNums[ i ]; j++ )
            curInput[i][j] = 0;

     initNodeLists();
     for( int i = 0; i < MAX_NODE_NUM; i++ )
          initNodeVariables( i );

     for( int i = 0; i < MAX_NODE_NUM; i++ )
          if( nodes[i] != null )
              nodes[i].reinit( 0, (Vector)null );
  }

  // test if the chunk is available
  public boolean isAvailableChunk( int nodeIdx )
  {
     if( nodes[ nodeIdx ] == null )
         return false;
     return nodes[ nodeIdx ].getAvailableFlag();
  }

/******************************** GET METHODS **********************************/

  public void getResults( int[] arr )
  {
     for( int i = 0; i < arr.length; i++ )
           arr[ i ] = withinReasoningResults[ i ];
  }

  public int getChosenChunkIdx()
  {
     return chosenNodeIdx;
  }

  public double getChosenChunkBLA()
  {
     return nodes[ chosenNodeIdx ].getBLA();
  }

  public double getChosenChunkActivation()
  {
     return nodes[ chosenNodeIdx ].getActivation();
  }

  public double getChosenChunkStrength()
  {
     return Math.max( C14 * typeStrengths[ chosenNodeIdx ][ RULE_BASED ],
                      C15 * typeStrengths[ chosenNodeIdx ][ SIMI_BASED ] );
  }

  public void getConclusionStrengths( double[] arr )
  {
     for( int i = 0; i < concChunkIndices.length; i++ )
     {
          short idx = concChunkIndices[ i ];
          arr[ i ] =  Math.max( C14 * typeStrengths[ idx ][ RULE_BASED ],
                                C15 * typeStrengths[ idx ][ SIMI_BASED ] );
     }
  }

  public void getChosenChunkContent( short[][] arr )
  {
     nodes[ chosenNodeIdx ].getContent( arr );
  }

  public GKChunk getChosenChunk()
  {
     return nodes[ chosenNodeIdx ];
  }

  public int getChunkIdx( String label )
  {
     for( int i = 0; i < availableNodeList.size(); i++ )
     {
          int idx = ((Integer)availableNodeList.elementAt( i )).intValue();
          if( label.compareTo( nodes[ idx ].getLabel() ) == 0 )
              return idx;
     }

     return -1;
  }

  public int getRuleNum( int idx )
  {
      int num = 0;
      for( int j = 0; j < ROW_NUM; j++ )
      {
           if( fromLinkNums[ idx ][ j ] > 0 )
               num += fromLinkNums[ idx ][ j ];
      }
      return num;
  }

  public String getRules( int idx, boolean flag )
  {
      String rule = "";
      String label = nodes[ idx ].getLabel();
      for( int j = 0; j < ROW_NUM; j++ )
      {
           if( fromLinkNums[ idx ][ j ] > 0 )
           {
               for( int k = 0; k < fromLinkNums[ idx ][ j ]; k++ )
               {
                    short idx1 = fromLinks[ idx ][ j ][ k ];
                    rule += label + " ----> " + nodes[ idx1 ].getLabel();
                    if( flag )
                        rule += " with age = " + String.valueOf( fromLinkAges[ idx ][ j ][ k ] );
                    rule += "\n";
               }
           }
      }

      return rule;
  }

  public String getChunk( int idx )
  {
      return nodes[ idx ].toString();
  }

/***************************** SET METHODS ***************************************/

  public void setCurrentTime( int time )
  {
     currentTime = time;
  }

  public void setConclusionDimIdx( int idx )
  {
     concDimIdx = (short)idx;
  }

  public void setConclusionIndices( short[] arr )
  {
     concChunkIndices = new short[ arr.length ];
     for( int i = 0; i < arr.length; i++ )
          concChunkIndices[ i ] = arr[ i ];
  }

  public void setMaxAges( int mNodeAge, int mLinkAge )
  {
     MAX_NODE_AGE = mNodeAge;
     MAX_LINK_AGE = mLinkAge;
  }

  public void setCurInput( int[] input )
  {
     int offset = 0;
     for( int i = 0; i < DIM_NUM; i++ )
     {
          int len = curInput[ i ].length;
          for( int j = 0; j < len; j++ )
               curInput[ i ][ j ] = input[ offset + j ];
          offset += len;
     }
  }

  public void setCueDimIndices( int[] indices )
  {
     cueDimNum = (short)indices.length;
     cueDimIndices = new short[ cueDimNum ];
     for( int i = 0; i < cueDimNum; i++ )
          cueDimIndices[ i ] = (short)indices[ i ];
  }

  public void setActivationThreshold( double val )
  {
     ACTIVATE_THRESHOLD = val;
  }

  public void setStrengthThreshold( double val )
  {
     STRENTH_THRESHOLD = val;
  }

  public void setSimilarityThreshold( double val )
  {
     SIMI_THRESHOLD = val;
  }

  public void setReasoningParas( double[] vals )
  {
     C14 = vals[ 0 ];
     C15 = vals[ 1 ];
  }

  public void setRetrievalCue( Vector c )
  {
     cue.removeAllElements();
     for( int i = 0; i < c.size(); i++ )
     {
          Object obj = c.elementAt( i );
          cue.insertElementAt( obj, i );
     }
  }

  public void setRetrievalCue( short[][] content )
  {
     cue.removeAllElements();
     for( int i = 0; i < content.length; i++ )
          cue.insertElementAt( content[ i ], i );
  }

  public String showLinks()
  {
     String str = "";

     str += "ALL of the \' from links \' \n";
     for( int i = 0; i < MAX_NODE_NUM; i++ )
     {
          str += "From node = " + String.valueOf( i ) + ",  links = ";
          for( int j = 0; j < ROW_NUM; j++ )
            for( int k = 0; k < fromLinkNums[ i ][ j ]; k++ )
                 str += "( " + String.valueOf( i ) + ", " + String.valueOf( fromLinks[i][j][k] ) + " ) ";
         str += "\n";
      }

     str += "ALL of the \' to links \' \n";
     for( int i = 0; i < MAX_NODE_NUM; i++ )
     {
          str += "To node = " + String.valueOf( i ) + ",  links = ";
          for( int j = 0; j < ROW_NUM; j++ )
            for( int k = 0; k < toLinkNums[ i ][ j ]; k++ )
                 str += "( " + String.valueOf( toLinks[i][j][k] ) + ", " + String.valueOf( i ) + " ) ";
          str += "\n";
     }

     return str;
  }

  public boolean checkLinkExist( short from, short to )
  {
     short slot = (short)(to / COLUMN_NUM);
     short num = fromLinkNums[ from ][ slot ];
     for( int i = 0; i < num; i++ )
     {
          if( fromLinks[ from ][ slot ][ i ] == to )
              return true;
     }

     return false;
  }

/**************************** MODIFIER METHODS **********************************/

  public int insertNode( int[] content, String label )
  {
     Vector con = new Vector( DIM_NUM );
     for( int i = 0; i < DIM_NUM; i++ )
     {
          short[] elem = new short[ dimValNums[i] ];
          for( int j = 0; j < dimValNums[i]; j++ )
               elem[ j ] = 0;
          int idx = content[ i ];
          elem[ idx ] = 1;
          con.addElement( elem );
     }

     int availableNodeIdx = 0;
     curNodeNum = availableNodeList.size();
     if( emptyNodeList.isEmpty() )
     {
         double[] inputs = new double[ MAX_NODE_NUM ];
         for( int i = 0; i < MAX_NODE_NUM; i++ )
              inputs[ i ] = ( (double)nodeAges[ i ] ) / SCALE_VAL;

         availableNodeIdx = nodeDecider.getRecommendation( inputs );
         nodes[ availableNodeIdx ].reinit( 0, con );
     }
     else
     {
         availableNodeIdx = (( Integer )( emptyNodeList.remove( 0 ) )).intValue();
         availableNodeList.addElement( new Integer(availableNodeIdx ) );
         if( nodes[ availableNodeIdx ] == null )
             nodes[ availableNodeIdx ] = new GKChunk( curNodeNum, 0, DIM_NUM, dimValNums,
                                                      dimTypes, INIT_BLA, weights );

         nodes[ availableNodeIdx ].reinit( 0, con );
     }

     nodes[ availableNodeIdx ].setLabel( label );
     nodes[ availableNodeIdx ].setAvailableFlag( true );
     initNodeVariables( availableNodeIdx );
     return availableNodeIdx;
  }

  public int insertNode( short[][] content, String label  )
  {
     Vector con = new Vector( DIM_NUM );
     for( int i = 0; i < DIM_NUM; i++ )
     {
          short[] elem = new short[ dimValNums[i] ];
          for( int j = 0; j < dimValNums[i]; j++ )
               elem[ j ] = content[ i ][ j ];
          con.addElement( elem );
     }

     int availableNodeIdx = 0;
     curNodeNum = availableNodeList.size();
     if( emptyNodeList.isEmpty() )
     {
         double[] inputs = new double[ MAX_NODE_NUM ];
         for( int i = 0; i < MAX_NODE_NUM; i++ )
              inputs[ i ] = ( (double)nodeAges[ i ] ) / SCALE_VAL;

         availableNodeIdx = nodeDecider.getRecommendation( inputs );
         nodes[ availableNodeIdx ].reinit( 0, con );
     }
     else
     {
         availableNodeIdx = (( Integer )( emptyNodeList.remove( 0 ) )).intValue();
         availableNodeList.addElement( new Integer(availableNodeIdx ) );
         if( nodes[ availableNodeIdx ] == null )
             nodes[ availableNodeIdx ] = new GKChunk( curNodeNum, 0, DIM_NUM, dimValNums,
                                                      dimTypes, INIT_BLA, weights );
         nodes[ availableNodeIdx ].reinit( 0, con );
     }

     nodes[ availableNodeIdx ].setLabel( label );
     nodes[ availableNodeIdx ].setAvailableFlag( true );
     initNodeVariables( availableNodeIdx );
     return availableNodeIdx;
  }

  public int insertNode( Vector con, String label )
  {
     int availableNodeIdx = 0;
     curNodeNum = availableNodeList.size();
     if( emptyNodeList.isEmpty() )
     {
         double[] inputs = new double[ MAX_NODE_NUM ];
         for( int i = 0; i < MAX_NODE_NUM; i++ )
              inputs[ i ] = ( (double)nodeAges[ i ] ) / SCALE_VAL;

         availableNodeIdx = nodeDecider.getRecommendation( inputs );
         nodes[ availableNodeIdx ].reinit( 0, con );
     }
     else
     {
         availableNodeIdx = (( Integer )( emptyNodeList.remove( 0 ) )).intValue();
         availableNodeList.addElement( new Integer(availableNodeIdx ) );
         if( nodes[ availableNodeIdx ] == null )
             nodes[ availableNodeIdx ] = new GKChunk( curNodeNum, 0, DIM_NUM, dimValNums,
                                                      dimTypes, INIT_BLA, weights );
         nodes[ availableNodeIdx ].reinit( 0, con );
     }

     nodes[ availableNodeIdx ].setLabel( label );
     nodes[ availableNodeIdx ].setAvailableFlag( true );
     initNodeVariables( availableNodeIdx );
     return availableNodeIdx;
  }

  public void removeNode( int nodeIdx )
  {
     Integer elem = new Integer( nodeIdx );
     availableNodeList.removeElement( elem );
     emptyNodeList.addElement( elem );

     nodes[ nodeIdx ].setLabel( "" );
     nodes[ nodeIdx ].setAvailableFlag( false );
     initNodeVariables( nodeIdx );
  }

  public void insertLink( short from, short to )
  {
     short slot, idx;

     slot = (short)(to / COLUMN_NUM);
     idx = fromLinkNums[ from ][ slot ];
     fromLinks[ from ][ slot ][ idx ] = to;
     fromLinkAges[ from ][ slot ][ idx ] = 0;
     fromLinkNums[ from ][ slot ]++;

     slot = (short)(from / COLUMN_NUM);
     idx = toLinkNums[ to ][ slot ];
     toLinks[ to ][ slot ][ idx ] = from;
     toLinkAges[ to ][ slot ][ idx ] = 0;
     toLinkNums[ to ][ slot ]++;
  }

  public void removeLink( short from, short to )
  {
     short idx, slot, len;

     idx = 0;
     slot = (short)(to / COLUMN_NUM);
     len = fromLinkNums[ from ][ slot ];

     while( idx < len )
     {
            if( fromLinks[ from ][ slot ][ idx ] != to )
                idx++;
            else
                break;
     }

     if( idx == len )
         return;

     for( int i = idx; i < len - 1; i++ )
     {
          fromLinks[from][slot][i] = fromLinks[from][slot][i+1];
          fromLinkAges[from][slot][i] = fromLinkAges[from][slot][i+1];
     }

     fromLinks[from][slot][ len-1 ] = EMPTY_LINK;
     fromLinkAges[from][slot][ len-1 ] = 0;
     fromLinkNums[ from ][ slot ]--;

     idx = 0;
     slot = (short)(from / COLUMN_NUM);
     len = toLinkNums[ to ][ slot ];

     while( idx < len )
     {
            if( toLinks[ to ][ slot ][ idx ] != from )
                idx++;
            else
                break;
     }

     if( idx == len )
         return;

     for( int i = idx; i < len - 1; i++ )
     {
          toLinks[to][slot][i] = toLinks[to][slot][i+1];
          toLinkAges[to][slot][i] = toLinkAges[to][slot][i+1];
     }

     toLinks[to][slot][ len-1 ] = EMPTY_LINK;
     toLinkAges[to][slot][ len-1 ] = 0;
     toLinkNums[ to ][ slot ]--;
  }

  /*
  public boolean isEqual( int idx, short[][] arr )
  {
     return nodes[ idx ].isEqual( arr );
  }
  */

  public void resetNodeAge( short idx )
  {
      nodeAges[ idx ] = 0;
  }

  public void resetLinkAge( short from, short to )
  {
     short slot = (short)(to / COLUMN_NUM);
     int len = fromLinkNums[ from ][ slot ];
     int idx = 0;
     while( idx < len )
     {
            if( fromLinks[ from ][ slot ][ idx ] != to )
                idx++;
            else
                break;
     }
     if( idx < len )
         fromLinkAges[ from ][ slot ][ idx ] = 0;

     slot = (short)(from / COLUMN_NUM);
     len = toLinkNums[ to ][ slot ];
     idx = 0;
     while( idx < len )
     {
            if( toLinks[ to ][ slot ][ idx ] != from )
                idx++;
            else
                break;
     }
     if( idx < len )
         toLinkAges[ to ][ slot ][ idx ] = 0;
  }

/***************************** REASONING METHODS *******************************/
  public boolean reasoningWithinNode()
  {
     curNodeNum = availableNodeList.size();
     if( curNodeNum == 0 )
         return false;

     if( reasoningType == RULE_BASED || reasoningType == RULE_SIMI_BASED )
         matchCue( RULE_BASED );
     if( reasoningType == SIMI_BASED || reasoningType == RULE_SIMI_BASED )
         matchCue( SIMI_BASED );

     activatedNodeNum = 0;
     for( int i = 0; i < MAX_NODE_NUM; i++ )
     {
          double actVal = Math.max( C14 * typeActivations[ i ][ RULE_BASED ],
                                    C15 * typeActivations[ i ][ SIMI_BASED ] );
          if( actVal > ACTIVATE_THRESHOLD )
          {
              activations[ activatedNodeNum ] = actVal;
              activatedNodeIndices[ activatedNodeNum++ ] = (short)i;
          }
     }

     if( activatedNodeNum == 0 )
         return false;

     withinReasoningResults = new short[ dimValNums[concDimIdx] ];

     int idx = nodeDecider.getRecommendation( activations, activatedNodeNum );
     chosenNodeIdx = activatedNodeIndices[ idx ];

     nodes[ chosenNodeIdx ].setConclusionDimIdx( concDimIdx );
     nodes[ chosenNodeIdx ].getConclusion( withinReasoningResults );
     return true;
  }

  public boolean reasoningBetweenNode()
  {
     curNodeNum = availableNodeList.size();
     if( curNodeNum == 0 )
         return false;

     if( reasoningType == RULE_BASED || reasoningType == RULE_SIMI_BASED )
     {
         matchCue( RULE_BASED );
         calcConcStrengths( RULE_BASED );
     }

     if( reasoningType == SIMI_BASED || reasoningType == RULE_SIMI_BASED )
     {
         matchCue( SIMI_BASED );
         calcConcStrengths( SIMI_BASED );
     }

     activatedNodeNum = 0;
     for( int i = 0; i < MAX_NODE_NUM; i++ )
     {
          double strength = Math.max( C14 * typeStrengths[ i ][ RULE_BASED ],
                                      C15 *  typeStrengths[ i ][ SIMI_BASED ] );
          if( strength > STRENTH_THRESHOLD )
          {
              strengths[ activatedNodeNum ] = strength;
              activatedNodeIndices[ activatedNodeNum++ ] = (short)i;
          }
     }

     if( activatedNodeNum == 0 )
         return false;

     int idx = nodeDecider.getRecommendation( strengths, activatedNodeNum );
     chosenNodeIdx = activatedNodeIndices[ idx ];
     nodes[ chosenNodeIdx ].getContent();

     return true;
  }

  public boolean reasoningBetweenNode( short[][] condition, short[][] conclusion )
  {
     curNodeNum = availableNodeList.size();
     if( curNodeNum == 0 )
         return false;

     conditionChunkIdx = -1;
     for( int i = 0; i < curNodeNum; i++ )
     {
          int idx = ( (Integer)availableNodeList.elementAt( i ) ).intValue();
          if( nodes[ idx ].isEqual( condition ) )
          {
              conditionChunkIdx = (short)idx;
              break;
          }
     }

     if( conditionChunkIdx == -1 )
         return false;

     conclusionChunkIdx = -1;
     for( int i = 0; i < curNodeNum; i++ )
     {
          int idx = ( (Integer)availableNodeList.elementAt( i ) ).intValue();
          if( nodes[ idx ].isEqual( conclusion ) )
          {
              conclusionChunkIdx = (short)idx;
              break;
          }
     }
     if( conclusionChunkIdx == -1 )
         return false;

     typeActivatedNodeNums[ RULE_BASED ] = 0;
     typeActivatedNodeNums[ SIMI_BASED ] = 0;
     for( int i = 0; i < MAX_NODE_NUM; i++ )
       for( int j = 0; j < 2; j++ )
       {
            typeActivatedNodeIndices[ i ][ j ] = 0;
            typeActivations[ i ][ j ] = 0.0;
       }

     if( reasoningType == RULE_BASED || reasoningType == RULE_SIMI_BASED )
         matchNode( conditionChunkIdx, RULE_BASED );
     if( reasoningType == SIMI_BASED || reasoningType == RULE_SIMI_BASED )
         matchNode( conditionChunkIdx, SIMI_BASED );

     chosenNodeIdx = conclusionChunkIdx;
     double linkActivation = calcLinkActivation( conditionChunkIdx, conclusionChunkIdx );
     for( int i = 0; i < 2; i++ )
          typeStrengths[ conclusionChunkIdx ][ i ] = linkActivation *
                                                    typeActivations[ conditionChunkIdx ][ i ];
     return true;
  }

/******************************* UPDATE METHODS ********************************/
  // used for the case: no retrieval is available.
  public void update()
  {
     curNodeNum = availableNodeList.size();
     if( curNodeNum == 0 )
     {
         System.out.println( "No available chunks!" );
         return;
     }

     for( int i = 0; i < curNodeNum; i++ )
     {
          int idx = ( (Integer)availableNodeList.elementAt( i ) ).intValue();
          nodeAges[ idx ]++;
     }

     updateLinkAges();
     checkNodeAges();
     checkLinkAges();
  }

  // used for the case: some retrieval is available.
  public void update( boolean flag )
  {
     nodes[ chosenNodeIdx ].increaseUsage(); // used for calculating BLA of chunks.

     curNodeNum = availableNodeList.size();
     for( int i = 0; i < curNodeNum; i++ )
     {
          int idx = ( (Integer)availableNodeList.elementAt( i ) ).intValue();
          if( idx != chosenNodeIdx )
              nodeAges[ idx ]++;
          else
              nodeAges[ idx ] = 0;
     }

     updateLinkAges();
     checkNodeAges();
     checkLinkAges();
  }

  // used for the case: some chunk should be extracted if not exists.
  public void update( int[] content, String label )
  {
     //modified on 02/14/02, no proper chunk has existed so far, create it.
     boolean bVal = checkExist( content );
     if( !bVal )
         insertNode( content, label );
     else
         nodes[ chosenNodeIdx ].increaseUsage(); // used for calculating BLA of chunks.

     curNodeNum = availableNodeList.size();
     for( int i = 0; i < curNodeNum; i++ )
     {
          int idx = ( (Integer)availableNodeList.elementAt( i ) ).intValue();
          nodeAges[ idx ]++;
     }

     updateLinkAges();
     checkNodeAges();
     checkLinkAges();
  }

/**************************** OVERRIDED METHODS *****************************/

  public String toString()
  {
     curNodeNum = availableNodeList.size();

     String str = "GKN( chunkNum = " + curNodeNum + " ) = \n";
     str += "{ ";
     for( int i = 0; i < curNodeNum; i++ )
     {
          int idx = ( (Integer)availableNodeList.elementAt( i ) ).intValue();
          str += nodes[ idx ];
     }
     str += " } \n";

     return str;
  }

/**************************** OTHER METHODS *****************************/

  public void calcChosenChunkBLA()
  {
     nodes[ chosenNodeIdx ].setCurrentTime( currentTime );
     nodes[ chosenNodeIdx ].calcBLA();
  }

/**************************** PRIVATE METHODS *****************************/

  private void checkNodeAges()
  {
     curNodeNum = availableNodeList.size();
     if( curNodeNum == 0 )
         return;

     for( int i = 0; i < availableNodeList.size(); i++ )
     {
          int idx = ( (Integer)availableNodeList.elementAt( i ) ).intValue();
          if( nodeAges[ idx ] >= MAX_NODE_AGE )
              removeNode( idx );
     }
  }

  private void checkLinkAges()
  {
     curNodeNum = availableNodeList.size();
     if( curNodeNum == 0 )
         return;

     for( int i = 0; i < curNodeNum; i++ )
     {
          int idx = ( (Integer)availableNodeList.elementAt( i ) ).intValue();
          for( int j = 0; j < ROW_NUM; j++ )
               if( fromLinkNums[idx][j] > 0 )
               {
                   int count = 0;
                   int len = fromLinkNums[idx][j];
                   short[] delArr = new short[ len ];

                   for( int k = 0; k < len; k++ )
                        if( fromLinkAges[idx][j][k] >= MAX_LINK_AGE )
                            delArr[ count++ ] = fromLinks[idx][j][k];

                   for( int k = 0; k < count; k++ )
                        removeLink( (short)idx, delArr[ k ] );
              }
    }
  }

  private void updateLinkAges()
  {
     curNodeNum = availableNodeList.size();
     if( curNodeNum == 0 )
         return;

     for( int i = 0; i < curNodeNum; i++ )
     {
          int idx = ( (Integer)availableNodeList.elementAt( i ) ).intValue();
          for( int j = 0; j < ROW_NUM; j++ )
          {
               if( fromLinkNums[idx][j] > 0 )
               {
                   int len = fromLinkNums[idx][j];
                   for( int k = 0; k < len; k++ )
                        fromLinkAges[idx][j][k]++;
               }

               if( toLinkNums[idx][j] > 0 )
               {
                   int len = toLinkNums[idx][j];
                   for( int k = 0; k < len; k++ )
                        toLinkAges[idx][j][k]++;
               }
          }
     }
  }

  private void matchCue( int type )
  {
     typeActivatedNodeNums[ type ] = 0;
     for( int i = 0; i < MAX_NODE_NUM; i++ )
     {
          typeActivatedNodeIndices[ i ][ type ] = 0;
          typeActivations[ i ][ type ] = 0.0;
     }

     curNodeNum = availableNodeList.size();
     for( int i = 0; i < curNodeNum; i++ )
     {
          int idx = ( (Integer)availableNodeList.elementAt( i ) ).intValue();
          matchNode( idx, type );
     }
  }

  private void matchNode( int idx, int type )
  {
      boolean bVal = false;
      nodes[ idx ].setCueDimIndices( cueDimIndices );
      if( type == RULE_BASED )
          bVal = nodes[ idx ].matchCueByRule( cue );
      if( type == SIMI_BASED )
          bVal = nodes[ idx ].matchCueBySimilarity( cue );
      nodes[ idx ].calcActivation( type );
      if( bVal )
      {
          double val =  nodes[ idx ].getActivation();
          typeActivatedNodeIndices[ typeActivatedNodeNums[ type ] ][ type ] = (short)idx;
          typeActivatedNodeNums[ type ]++;
          typeActivations[ idx ][ type ] = val;
      }
  }

  private void calcConcStrengths( int type )
  {
     for( int i = 0; i < MAX_NODE_NUM; i++ )
          typeStrengths[ i ][ type ] = 0.0;

     double strength = 0.0;
     for( int i = 0; i < typeActivatedNodeNums[ type ]; i++ )
     {
          short idx = typeActivatedNodeIndices[ i ][ type ];
          double actVal = nodes[ idx ].getActivation();
          for( int j = 0; j < ROW_NUM; j++ )
          {
               if( fromLinkNums[ idx ][ j ] > 0 )
               {
                   for( int k = 0; k < fromLinkNums[ idx ][ j ]; k++ )
                   {
                        short idx1 = fromLinks[ idx ][ j ][ k ];
                        strength = typeStrengths[ idx1 ][ type ];
                        if( actVal > strength )
                            typeStrengths[ idx1 ][ type ] = actVal;
                   }
               }
           }
      }
  }

  private double calcLinkActivation( int from, int to )
  {
     short slot = (short)(to / COLUMN_NUM);
     short idx = fromLinkNums[ from ][ slot ];
     for( int i = 0; i < idx; i++ )
        if( fromLinks[ from ][ slot ][ i ] == to )
            return 1.0;
      return 0.0;
  }

  private boolean checkExist( int[] content )
  {
     boolean bVal = false;

     curNodeNum = availableNodeList.size();
     if( curNodeNum == 0 )
         return bVal;

     for( int i = 0; i < curNodeNum; i++ )
     {
          int idx = ( (Integer)availableNodeList.elementAt( i ) ).intValue();
          bVal = nodes[ idx ].contain( content );
          if( bVal )
          {
              chosenNodeIdx = idx;
              break;
          }
     }

     return bVal;
  }
}