package clarion.nacs;
import java.util.Vector;
import java.util.Arrays;
import clarion.common.*;
import clarion.tools.*;
import clarion.system.*;

public class GKSChunk extends Chunk {

   final int AMN_IDX = 0, GKS_IDX = 1;

   // global setting variable.
   Global global;
   int    idnIdx = -1;    // the IDN index the chunk from
   int    amnIdx = -1;    // the AMN index the chunk in
   int    chunkID;        // id of the chunk.
   int    age;            // age from last usage.
   int    retrievalTime;  // chunk retrieval time.
   long   currentTime;
   short  CONC_IDX;       // the conclusion dimension.
   AssocRuleGrp rules;    // an associative rule group formed by same conclusion.
   int    activationMode; // mode of activation: minimum or average.
   short  dimMatches[];   // record the matched dimensions with the cue.
   double weights[];      //
   double dimStrengths[]; // dimensional strength.
   boolean inEmOn;
   boolean availableFlag;

   // new constructor.
   public GKSChunk( int cId, int type, String label, Feature[] features,
                    int idnIdx, int amnIdx, Global g )
   {
      super( type, label, features, g );
      chunkID = cId;
      this.idnIdx = idnIdx;
      this.amnIdx = amnIdx;
      global = g;
      rules = new AssocRuleGrp( chunkID, global );
      activationMode = global.getActivationMeasure();
      dimMatches = new short[ getDimNum() ];
      weights = global.getTaskClarion().getAssocDimWeights( getDimNum() );
      dimStrengths = new double[ getDimNum() ];
   }

   public GKSChunk( int cId, int type, String label, Chunk[] chunks,
                    int idnIdx, int amnIdx, Global g )
   {
      super( type, label, chunks, g );
      chunkID = cId;
      this.idnIdx = idnIdx;
      this.amnIdx = amnIdx;
      global = g;
      rules = new AssocRuleGrp( chunkID, global );
      activationMode = global.getActivationMeasure();
      dimMatches = new short[ getDimNum() ];
      dimStrengths = new double[ getDimNum() ];
   }

/***************************** GET AND SET METHODS ****************************/
   public int getIdnIdx()
   {
      return idnIdx;
   }

   public int getAmnIdx()
   {
      return amnIdx;
   }

   public int getAge()
   {
      return age;
   }

   public int getRuleNum()
   {
      return rules.getRuleNum();
   }

   public AssocRuleGrp getRules()
   {
      return rules;
   }

   public short[][] getActiveRuleConditions()
   {
      return rules.getActiveRuleConditions();
   }

   public int getConclusionDimIdx( )
   {
      return CONC_IDX;
   }

   public boolean getAvailableFlag()
   {
      return availableFlag;
   }

   public boolean isInEmOn()
   {
      return inEmOn;
   }

   public Feature getConclusion()
   {
      return featList[ CONC_IDX ];
   }

   public int getRetrievalTime()
   {
      return retrievalTime;
   }

   public int getAssocRuleAppTime()
   {
      return rules.getAssocRuleAppTime();
   }

   public void resetAge()
   {
      age = 0;
   }

   public void setCurrentTime( long curTime )
   {
      currentTime = curTime;
      rules.setCurrentTime( currentTime );
   }

   public void setRules( AssocRuleGrp ruleGrp )
   {
      try{
          if( ruleGrp != null )
              rules = (AssocRuleGrp)(ruleGrp.clone());
      }
      catch(  CloneNotSupportedException e ) {
         System.out.println( "clone AssocRuleGrp error from GKSChunk.java" );
      }
   }

   public void setAvailableFlag( boolean flag )
   {
      availableFlag = flag;
   }

   public void setInEmOn( boolean flag )
   {
      inEmOn = flag;
   }

   public void setConclusionDimIdx( short idx )
   {
       CONC_IDX = idx;
   }

/*************************** REASONING METHODS ********************************/
   // to check if this chunk is activated. To say a chunk is activated, it means
   // that all or some of its dimensions are activated to a strength level.
   // A dimension of the chunk is activated to a strength level means one of
   // allowable values in the dimension is activated with a strength( only for
   // the case that the value # cue has in that dimension is 1. )
   // Generally, To say a dimension of a chunk is activated, all of the values
   // cue has in the dimension should be contained in the dimension of the chunk.
   public boolean activateChunk( Feature[] cue, int whichLevel )
   {
       for( int i = 0; i < getDimNum(); i++ ){
            dimMatches[ i ] = 0;
            dimStrengths[ i ] = 0.0;
       }
       // fill in dimensional strengths.
       int matchCount = 0;
       if( featList.length > 0 ){
           for( int i = 0; i < cue.length; i++ ){
                int idx = searchFeature( cue[ i ].getLabel() );
                if( idx == -1 || cue[ i ].getType() != featList[ idx ].getType() )
                    continue;
                Feature f1 = featList[ idx ];
                Feature f2 = cue[ i ];
                dimMatches[ idx ] = CommonMethods.contain( f1.getContent(), f2.getContent() );
                if( dimMatches[ idx ] == 1 ){
                    matchCount++;
                    if( whichLevel == GKS_IDX )
                        dimStrengths[ idx ] = 1.0;
                    else
                        dimStrengths[ idx ] = calcDimStrength( f1, f2 );
                }
                else
                    dimStrengths[ idx ] = 0.0;
            }
       }
       // calculate strength.
       strength = 0.0;
       if( matchCount == 0 || (!global.isAssocPartMatchOn() && matchCount < getDimNum()) )
           return false;
       if( whichLevel == GKS_IDX )
           for( int i = 0; i < getDimNum(); i++ )
                strength += dimStrengths[i] * weights[i];
       else
           for( int i = 0; i < getDimNum(); i++ )
                strength += dimStrengths[i] / getDimNum() ;
       return true;
   }

   // calculate the support from the associative rules pointing to this chunk.
   public void calculateStrength( Vector con )
   {
       if( getRuleNum() > 0 )
           strength = rules.drawConclusion( con );
   }

   // search a rule with the condition.
   public boolean searchRule( short[] conditions )
   {
       return rules.searchRule( conditions );
   }

/****************************** BLA METHODS ***********************************/
   public void increaseUsage()
   {
       super.increaseUsage( currentTime );
   }

   public void calcRetrievalTime()
   {
      /*
      double t2 = global.getChunkRetTime_t2();
      double t3 = global.getChunkRetTime_t3();
      super.calcBLA( currentTime );
      retrievalTime = (int)(t2 + t3 / BLA);
      */
      super.calcBLA( currentTime );
      retrievalTime = global.getTaskClarion().getChunkRetrTime( BLA );
   }

/************************ INSERT AND REMOVE METHODS ***************************/
   public boolean insertRule( short[] condition, short conclusion )
   {
       return rules.insertRule( condition, conclusion );
   }

   public boolean removeRule( short[] condition, short conclusion )
   {
       return rules.removeRule( condition, conclusion );
   }

   public boolean removeCondition( int chunkIdx )
   {
       return rules.removeCondition( chunkIdx );
   }

/*************************** CHUNK UPDATE METHODS *****************************/
   public boolean update()
   {
       age++;
       int MAX_CHUNK_AGE = (int)( 1.0 / global.getChunkDensity() );
       if( age >= MAX_CHUNK_AGE ){
           //System.out.println( "age = " + age );
           return true;
       }

       int MAX_RULE_AGE = (int)( 1.0 / global.getAssocDensity() );
       rules.setMaxRuleAge( MAX_RULE_AGE );
       rules.update();

       return false;
   }

/************************** OVERRIDED METHODS ********************************/
   public String toString()
   {
      String str = super.toString();
      str += "CHUNK AGE = " + String.valueOf( age ) + "\n";
      str += "RULES-IN = \n" + rules;
      return str;
   }

   public Object clone() throws CloneNotSupportedException
   {
       GKSChunk chunk = new GKSChunk( chunkID, chunkType, chunkLabel, featList, idnIdx, amnIdx, global );
       // from super class.
       chunk.setStrength( strength );
       chunk.setBlaOption( chunkBlaOption );
       if( chunkBlaOption == global.CONSTANT_OPTION )
           chunk.setBla( BLA );
       else{
           chunk.setBLA_C( c );
           chunk.setBLA_D( d );
           chunk.setBLA_INIT( iBLA );
           chunk.setUsageNum( usageNum );
           chunk.setUsageRecords( usageRecords );
       }
       // from this class
       chunk.setConclusionDimIdx( CONC_IDX );
       chunk.setAvailableFlag( availableFlag );
       chunk.setRules( rules );
       //chunk.setAge( age );
       return chunk;
   }

/************************* PRIVATE METHODS ************************************/
   //calculate f1 dimension strength of the chunk activated by AMN f2 dimension.
   private double calcDimStrength( Feature f1, Feature f2 )
   {
       double actVal = 0.0;
       short[] values = f1.getValues();
       if( activationMode == 0 ){ // minimum mode.
           actVal = 1.0;
           for( int i = 0; i < values.length; i++ ){
                double val = f2.getActivations()[ values[ i ] ];
                if( val < actVal )
                    actVal = val;
           }
       }
       else{  // average mode.
           for( int i = 0; i < values.length; i++ ){
                double val = f2.getActivations()[ values[ i ] ];
                actVal += val;
           }
           actVal /= values.length;
       }
       return actVal;
   }
}