package clarion.system;
import java.util.Vector;
import clarion.acs.*;
import clarion.nacs.*;
import clarion.acs.gs.*;
import clarion.acs.wm.*;
import clarion.mcs.*;
import clarion.ms.*;
import clarion.common.*;
import clarion.coordinate.*;
import clarion.tools.*;

public class Clarion {

   TaskAgent agent;          // current agent.
   Global  global;           // global settings.
   TaskClarion taskClarion;  // task-specific clarion settings.
   ACS     acs;              // action-centered subsystem.
   NACS    nacs;             // non-action-centered subsystem.
   MCS     mcs;
   MS      ms;
   Drives  drives;
   ExGScoordinator coordinator;
   int     activeAmnIdx;
   int     actionTypeNum;
   Feature sensoryInput[];   // sensory input.
   Feature nacsInput[];
   Feature desiredOutput[][];
   short[] chosenActions[];
   double  reinforcements[];
   GKSChunk retrievalResults[];

   public Clarion( Global g, TaskAgent taskAgent )
   {
      global = g;
      global.setClarion( this );
      agent = taskAgent;
      taskClarion = global.getTaskClarion();
      // nacs.
      if( global.isNacsOn() )
          nacs = new NACS( g );
      // acs.
      acs = new ACS( g );
      // initialize supervisory subsystem.
      if( global.isSupervisorySubsysOn() ){
          if( global.isMcsOn() ){
              drives = new Drives();
              mcs = new MCS( acs, nacs, drives, global );
          }
          if( global.isMsOn() )
              ms = new MS( global, global.getGS() );
      }
      // initialize coordinator.
      coordinator = new ExGScoordinator( agent, global );
      actionTypeNum = global.getAcsNetNum();
      chosenActions = new short[ actionTypeNum ][];
      for( int i = 0; i< actionTypeNum; i++ )
           chosenActions[i] = new short[ global.getIdnOutputDimNum(i) ];
      reinforcements = new double[ actionTypeNum ];
   }

/*************************** GET METHODS **************************************/
   public ACS getACS()
   {
      return acs;
   }

   public NACS getNACS()
   {
      return nacs;
   }

   public Feature[] getSensoryInput()
   {
      return sensoryInput;
   }

   public String[] getRerRules( int netIdx )
   {
      return acs.getRerRules( netIdx );
   }

   public String[] getIrlRules( int netIdx )
   {
      return acs.getIrlRules( netIdx );
   }

   public String[] getFixRules( int netIdx )
   {
      return acs.getFixRules( netIdx );
   }

   public int getRT()
   {
      int RT = 0;
      for( int i = 0; i < actionTypeNum; i++ ){
           if( chosenActions[i][0] != global.EMPTY && !(global.isWmNet(i)||global.isGsNet(i)))
           { // an external action.
               if( agent.isControlNacsAction( i, chosenActions[i] ) ) { // action for controlling NACS
                   RT = acs.getRT( nacs.getRT() );
               }
               else   // normal ACS action.
                   RT = acs.getRT();
           }
      }
      return RT;
   }

   public GKSChunk[] getRetrievalResults()
   {
      boolean retVal = this.isRetrievalAction();
      //System.out.println( "retrieval action = " + retVal );
      if( retVal ){
          return retrievalResults;
      }
      return null;
   }

   public void setCurrentTime( long curTime )
   {
      acs.setCurrentTime( curTime );
      if( global.isNacsOn() )
          nacs.setCurrentTime( curTime );
   }

   public void training()
   {
      sensoryInput = agent.getCurSensoryInput();
      acs.observeState( sensoryInput );
      if( !agent.isInLearningTest() && global.isNacsOn() ){
          nacsInput = agent.getCurNacsInput();
          nacs.observeState( nacsInput );
      }
      acs.reasoning();
      CommonMethods.resetArray( chosenActions, Global.EMPTY );
      acs.getChosenActions( chosenActions );
      performActions();
      // controlled by MCS, ( may be revised ).
      if( global.isMcsOn() )
          fillInMonitorBuffer();
      // update subsystems.
      sensoryInput = agent.getCurSensoryInput();
      desiredOutput = agent.getCurDesiredOutput();
      acs.setCurDesiredOutput( desiredOutput );
      if( !agent.isInLearningTest() ){
          acs.update( sensoryInput, reinforcements );
          if( global.isNacsOn() && isRetrievalAction() ){
              retrievalResults = nacs.getResults( );
              nacsInput = agent.getCurNacsInput();
              activeAmnIdx = agent.getActiveAmnIdx();
              nacs.update( activeAmnIdx, chosenActions, nacsInput, reinforcements );
          }
      }
   }

   public void testing()
   {
      sensoryInput = agent.getCurSensoryInput();
      acs.observeState( sensoryInput );
      if( !agent.isInLearningTest() && global.isNacsOn() ){
          nacsInput = agent.getCurNacsInput();
          nacs.observeState( nacsInput );
      }
      acs.reasoning();
      CommonMethods.resetArray( chosenActions, Global.EMPTY );
      acs.getChosenActions( chosenActions );
      performActions();
      // controlled by MCS, ( may be revised ).
      if( global.isMcsOn() )
          fillInMonitorBuffer();
      //sensoryInput = agent.getCurSensoryInput();
      //acs.update( sensoryInput, reinforcements );
      if( global.isNacsOn() ){
          retrievalResults = nacs.getResults();
          //nacsInput = agent.getCurNacsInput();
          //activeAmnIdx = agent.getActiveAmnIdx();
          //nacs.update( activeAmnIdx, chosenActions, nacsInput, reinforcements );
      }
   }

   public String testAem( int acsNetIdx )
   {
      return nacs.testAem( acsNetIdx );
   }

   // let all of the ACS networks perform their actions including EX,GS,WM if exist.
   public void performActions()
   {
      GKSChunk[] wmChunks = global.isWmOn() ? global.getWM().getItems():null;
      GoalChunk goal = global.getGsOption()!=Global.GS_NON ? global.getGS().getActiveItem() : null;
      //all ACS networks perform their actions including EX,GS,WM if exist.
      coordinator.setChosenActions( chosenActions );
      coordinator.coordinate();
      Feature[] newInput = agent.getCurSensoryInput();
      GKSChunk[] newWmChunks = global.isWmOn() ? global.getWM().getItems() : null;
      GoalChunk newGoal = global.getGsOption()!=Global.GS_NON ? global.getGS().getActiveItem() : null;
      for( int i = 0; i < actionTypeNum; i++ )
           if( chosenActions[i][0] != global.EMPTY ){
               reinforcements[i] = taskClarion.getReinforcement( i, sensoryInput, wmChunks, goal,
                                   chosenActions[i], newInput, newWmChunks, newGoal);
           }
   }

   private void fillInMonitorBuffer()
   {
      MonitorBuf buffer = global.getMCS().getMonitorBuffer();
      Chunk[][][] acsConclusions = acs.getAllConclusions();
      for( int i = 0; i < acsConclusions.length; i++ )
        for( int j = 0; j < acsConclusions[i].length; j++ )
             buffer.setCurAcsConclusions( i, j, acsConclusions[i][j] );
      if( global.isNacsOn() && isRetrievalAction() ){
          Chunk[][] nacsConclusions = nacs.getAllConclusions();
          for( int i = 0; i < nacsConclusions.length; i++ )
               buffer.setCurNacsConclusions( i, nacsConclusions[i] );
      }
   }

   private boolean isRetrievalAction()
   {
      boolean retVal = false;
      for( int i = 0; i < chosenActions.length; i++ )
           if( global.isExNet(i) && chosenActions[i][Global.D_TYPE] == 1 &&
               chosenActions[i][Global.D_ACT] == 1 && chosenActions[i][Global.D_INFER] == 1){
               retVal = true;
               break;
           }
      return retVal;
   }
}