package clarion.tools.linkedlist;

public class LinkedList implements List, Cloneable {


	//=================== ACCESSORS =======================


	/** Checks whether the list is empty.
	    @return <tt>true</tt> if the list is empty,
	            <tt>false</tt> otherwise. */
	//---------------------------------------------
	public boolean isEmpty () {
		return count==0;
	}

	/** Counts the number of elements in the list.
	    @return the number of elements in the list. */
	//---------------------------------------------
	public int getCount () {
		return count;
	}

	/** Returns the first element in the list.
	    @return the first element in the list. */
	//---------------------------------------------
	public Object getHead () {
		if(count==0) throw new java.util.NoSuchElementException();
		return head.element;
	}

	/** Returns the last element in the list.
	    @return the last element in the list. */
	//---------------------------------------------
	public Object getTail () {
		if(count==0) throw new java.util.NoSuchElementException();
		return tail.element;
	}

	/** Returns the element at the specified position in the list.
	    @param index the index of the element to return.
	    @return the element at the specified position. */
	//---------------------------------------------
	public Object getFrom (int index) {
		if(index<0 || index>=count) throw new IndexOutOfBoundsException();
		if(index==0) return getHead();
		if(index==count-1) return getTail();
		Node node=head;
		for(int i=0;i<index;i++) node=node.next;
		return node.element;
	}


	//=================== CANONICAL METHODS =======================


	/** Returns a string representation of the list.
	    @return a string representation of the list. */
	//---------------------------------------------
	public String toString () {
		String s="LinkedList[";
		int i=0;
		for(Node node=head;node!=null;node=node.next)
			{s+="("+i+","+node.element+")"; i++;}
		return s+"]";
	}

	/** Checks whether some other object is equal to this one.
		@param other the object with which to compare.
	    @return <tt>true</tt> if the two objects are equal,
	    		<tt>false</tt> otherwise. */
	//---------------------------------------------
	public boolean equals (Object other) {
		if(other ==null || !(other instanceof List)) return false;
		List otherList = (List) other;
		if(this.count!=otherList.getCount()) return false;
		for(int i=0;i<count;i++) {
			Object thisElement=this.getFrom(i);
			Object otherElement=otherList.getFrom(i);
			if(thisElement==null) {
				if(otherElement!=null) return false;
			}
			else
				if(!thisElement.equals(otherElement)) return false;
		}
		return true;
	}

	/** Computes a hash code value for the list.
	    @return a hash code value for the list. */
	//---------------------------------------------
	public int hashCode () {
		int sum=0;
		int i=0;
		Node node = head;
		while(i<4 && node!=null) {
			if(node.element!=null) {
				sum<<=8;
				sum|=node.element.hashCode()&0xFF;
				i++;
			}
			node=node.next;
		}
		return sum;
	}

	/** Creates a deep copy of <tt>this</tt> list.
	    @return a clone of <tt>this</tt> list.
	    @throws CloneNotSupportedException if some element in the list
	    		is not an instance of <tt>CloneableListElement</tt>. */
	//---------------------------------------------
	public Object clone () throws CloneNotSupportedException {
		LinkedList list = new LinkedList();
		for(Node node=head;node!=null;node=node.next) {
			Object e=null;
			if(node.element!=null) {
				if(!(node.element instanceof CloneableListElement))
					throw new CloneNotSupportedException();
				e=((CloneableListElement)node.element).clone();
			}
			list.insertTail(e);
		}
		return list;
	}


	//=================== MODIFIERS =======================


	/** Inserts an element at the beginning of the list.
	    @param element the element to be inserted. */
	//---------------------------------------------
	public void insertHead (Object element) {
		head=new Node(element,null,head);
		if(count==0) tail=head;
		else head.next.prev=head;
		count++;
	}

	/** Inserts an element at the end of the list.
	    @param element the element to be inserted. */
	//---------------------------------------------
	public void insertTail (Object element) {
		tail=new Node(element,tail,null);
		if(count==0) head=tail;
		else tail.prev.next=tail;
		count++;
	}

	/** Inserts an element at the specified position in the list.
	    @param element the element to be inserted.
	    @param index the position at which the element is to be inserted. */
	//---------------------------------------------
	public void insertAt (Object element, int index) {
		if(index<0 || index>count) throw new IndexOutOfBoundsException();
		if(index==0) insertHead(element);
		else if(index==count) insertTail(element);
		else {
			Node node=head;
			for(int i=0;i<index;i++) node=node.next;
			Node newNode=new Node(element,node.prev,node);
			node.prev.next=newNode;
			node.prev=newNode;
			count++;
		}
	}

	/** Removes the first element from the list.
	    @return the element that was first in the list. */
	//---------------------------------------------
	public Object removeHead () {
		if(count==0) throw new java.util.NoSuchElementException();
		Object e=head.element;
		head=head.next;
		if(count==1) tail=null;
		else head.prev=null;
		count--;
		return e;
	}

	/** Removes the last element from the list.
	    @return the element that was last in the list. */
	//---------------------------------------------
	public Object removeTail () {
		if(count==0) throw new java.util.NoSuchElementException();
		Object e=tail.element;
		tail=tail.prev;
		if(count==1) head=null;
		else tail.next=null;
		count--;
		return e;
	}

	/** Removes from the list the element at the specified location.
	    @param index the index of the element to be removed.
	    @return the element that was at the specified location. */
	//---------------------------------------------
	public Object removeFrom (int index) {
		if(index<0 || index>=count) throw new IndexOutOfBoundsException();
		if(index==0) return removeHead();
		else if(index==count-1) return removeTail();
		else {
			Node node=head;
			for(int i=0;i<index;i++) node=node.next;
			node.prev.next=node.next;
			node.next.prev=node.prev;
			count--;
			return node.element;
		}
	}


	//======================================================


	protected Node head, tail;

	protected int count;

	protected static class Node {
		Object element;
		Node next, prev;
		Node (Object e, Node p, Node n) {
			element=e;
			next=n;
			prev=p;
		}
	}
}
